' //////////////////////////////////////////////////////////////////////
' TITLE: EPM_Defnder_Sound_Engine2_xxx.spin
'
' DESCRIPTION:
'
' AUTHOR:  Eric Moyer, Copyright (C)2007
'
'
'  Rev  Date      Description
'  ---  --------  ---------------------------------------------------
'  001  12-01-07  Initial Creation (Split sound into two engines)
'                           
' //////////////////////////////////////////////////////////////////////
'
'  Audio channel usage:
'     0: Thrust
'     1: (All others)
'
' //////////////////////////////////////////////////////////////////////
CON

  CLOCK_RATE_HYDRA       = 80_000_000
  CLOCK_RATE_HYBRID      = 96_000_000
  PIN__AUDIO             = %00000000_00000000_00000000_10000000

  AUDIO_SAMPLE_RATE      = 22_000  'Audio generation rate

  '------------------------------------
  'HARDWARE IDs
  '------------------------------------
  HARDWARE_HYDRA         = 1 
  HARDWARE_HYBRID        = 2
  
  '------------------------------------
  'SOUND EFFECT IDs
  '------------------------------------

  'Internally generated IDs (do not use in START calls)
  SOUND__INT_EXPLODE     = %00000000_00000000_00000000_00000001 'Handles all "explosion" types
  SOUND__INT_IDLE_THRUST = %00000000_00000000_00000000_00000010 'Handles idle thrust sound

  'Internally masks (do not use in START calls)  
  MASK_EXPLOSIONS        = %00000000_00000000_00000000_01111100 'All explostion types
  MASK_UNSUPPORTED       = %00000000_00001110_00000000_00000000 'Sounds unsupported by this engine (i.e. sounds
                                                                ' handled by the other engine)
  'Exteranally accessible IDs
  SOUND__HUMAN_DIE       = %00000000_00000000_00000000_00000100 
  SOUND__FIRE            = %00000000_00000000_00000000_00001000 
  SOUND__WARP            = %00000000_00000000_00000000_00010000 
  SOUND__YOU_DIE         = %00000000_00000000_00000000_00100000 
  SOUND__PLANET_EXPLODE  = %00000000_00000000_00000000_01000000 

  SOUND__THRUST          = %00000000_00000000_00000000_10000000 
  SOUND__MUTANT          = %00000000_00000000_00000001_00000000 
  SOUND__LANDER_FIRE     = %00000000_00000000_00000010_00000000 
  SOUND__LANDER_DIE      = %00000000_00000000_00000100_00000000 
  SOUND__HUMAN_ABDUCTED  = %00000000_00000000_00001000_00000000 
  SOUND__HUMAN_FALLING   = %00000000_00000000_00010000_00000000 
  SOUND__HUMAN_CAUGHT    = %00000000_00000000_00100000_00000000 
  SOUND__HUMAN_RETURNED  = %00000000_00000000_01000000_00000000 
  SOUND__SWARMER         = %00000000_00000000_10000000_00000000 
  SOUND__EXTRA_LIFE      = %00000000_00000001_00000000_00000000 
  SOUND__LEVEL_DONE      = %00000000_00000010_00000000_00000000 
  SOUND__START_GAME      = %00000000_00000100_00000000_00000000 
  SOUND__HIGHSCORE       = %00000000_00001000_00000000_00000000 
  

  '------------------------------------
  'Flags
  '------------------------------------  
  FLAG_RISING_EDGE       = %00000001
  FLAG_FALLING_EDGE      = %00000010

  '------------------------------------
  'Misc
  '------------------------------------  
  MUTANT_SAMPLE_WIDTH_INIT = 2
  MUTANT_SAMPLES_PER_LWORD = 16

  LFIRE_SAMPLE_WIDTH_INIT = 3 
  LFIRE_SAMPLES_PER_LWORD = 32

  COG_STOPPED             = 99
  
OBJ
  'dbg       :       "PasDebug"                       '<---- Add for Debugger
  
VAR
   long               this_cog 

PUB init 
   this_cog := COG_STOPPED
  
PUB start (sfx_control_block_p)

  if this_cog == COG_STOPPED
    this_cog := cognew(@sfx_entry,sfx_control_block_p)
'   dbg.start(31,30,@sfx_entry)                 '<---- Add for Debugger   

PUB stop 
  if (this_cog <> COG_STOPPED)  
    cogstop(this_cog)
    this_cog := COG_STOPPED

'///////////////////////////////////////////////////////////////////////
' DATA SECTION /////////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

DAT
                        
'------------------------------------
'Entry
'------------------------------------
                        org        0
                        
sfx_entry

'  --------- Debugger Kernel add this at Entry (Addr 0) ---------
'   long $34FC1202,$6CE81201,$83C120B,$8BC0E0A,$E87C0E03,$8BC0E0A
'   long $EC7C0E05,$A0BC1207,$5C7C0003,$5C7C0003,$7FFC,$7FF8
'  -------------------------------------------------------------- 

                        'get pointers to parameters in the Effect Control block
                        mov     r1,PAR             'Get pointer to Effect Control block

                        'Get Pointer to sfx_start_flags
                        mov     p_sfx_start_flags, r1
                        add     r1, #4

                        'Get Pointer to sfx_stop_flags
                        mov     p_sfx_stop_flags, r1
                        add     r1, #4

                        'Get (and process) sfx_hardware_type
                        rdlong  r2, r1             ' Get hardware id
                        cmp     r2, #HARDWARE_HYDRA wz
              if_z      mov     sample_interval_ticks, SAMPLE_INTERVAL_HYDRA
              if_nz     mov     sample_interval_ticks, SAMPLE_INTERVAL_HYBRID  

'------------------------------------
'Effect processing loop
'------------------------------------

                        '------------------------------------
                        'Init 
                        '------------------------------------
                        mov     sfx_playing_mask, #0

                        'Configure audio pin
                        andn    outa,_PIN__AUDIO       ' Set audio pin LOW          "0"
                        or      dira,_PIN__AUDIO       ' Set audio pin to OUTPUT
                        mov     ctra, CTRAVAL
                        mov     frqa, LONG_HALF
                        mov     phsa, #$ff

                        'DEBUG LED
                        'or      dira, #1

                        'seed the LFSR pseudo-random generator
                        mov     random_number, #1  

                        mov     start_request_previous, #0
                        mov     start_request_current, #0
                        mov     start_request_actions, #0
                        mov     sfx_in_progress, #0
:restart                        
                        mov     next_sample_start_time, cnt
                        add     next_sample_start_time, sample_interval_ticks

                        mov     previous_pulse_level, #0


:sample_sync            'wait for next audio sample start time
                        waitcnt next_sample_start_time, sample_interval_ticks

                        
                        mov     flags, #0
                        mov     AUDIO_channel_0, #0
                        mov     AUDIO_channel_1, #0

                        '------------------------------------
                        'Generate LFSR Pseudo random number                                      
                        '------------------------------------
                        mov    x, #32                              'Generate 32 bits
:lfsr_loop              mov    y, random_number
                        and    y, #1
                        neg    y, y
                        and    y, LFSR_32_TAPS
          
                        xor    random_number, y
                        ror    random_number, #1
                        sub    x, #1  wz
              if_nz     jmp    #:lfsr_loop  
 
                        
                        '------------------------------------
                        'Process SFX start requests
                        '------------------------------------

                        'Determine if any SFX's have been requested to start
                        rdlong  start_request_current, p_sfx_start_flags       'get current start flags
                        mov     start_request_actions, start_request_current   'valid starts are any newly set bits
                        andn    start_request_actions, start_request_previous
                        mov     start_request_previous, start_request_current 'save current as previoius

                        '------------------------------------
                        'Process SFX stop requests
                        '------------------------------------
                        'Determine if any SFX's have been requested to stop
                        rdlong  stop_request_current, p_sfx_stop_flags         'get current stop flags
                        andn    sfx_in_progress, stop_request_current

                        
                        '------------------------------------
                        'START: Level Done 
                        '------------------------------------
                        mov     sfx_id_mask,SOUND__LEVEL_DONE_ID
                        test    start_request_actions, sfx_id_mask wz

                        'Configure leveldone pulse
               if_nz    mov     pulse1_init, LONG_HALF
               if_nz    sub     pulse1_init, PULSE1_INIT_START
               if_nz    mov     pulse1_base_init, PULSE1_INIT_START
               if_nz    mov     pulse1_end, LONG_HALF
               if_nz    add     pulse1_end, PULSE1_END_START
                        
               if_nz    mov     pulse2_init, LONG_HALF
               if_nz    sub     pulse2_init, PULSE2_INIT_START
               if_nz    mov     pulse2_end, LONG_HALF
               if_nz    add     pulse2_end, PULSE2_END_START

               if_nz    mov     pulse1_val, pulse1_init
               if_nz    mov     pulse2_val, pulse2_init 

                        'Indicate that the sound is in progress  
               if_nz    or     sfx_in_progress, sfx_id_mask

                        '------------------------------------
                        'START: Start Game 
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

               if_nz    mov     SGAME_angle, #0
               if_nz    mov     SGAME_scale, #1

                        'Indicate that the sound is in progress
               if_nz    or     sfx_in_progress, sfx_id_mask


                        '================================================================================
                        'PLAY: Level Done 
                        '================================================================================  
                        mov     sfx_id_mask, SOUND__LEVEL_DONE_ID        
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:ldone_bypass

                        
                        'Compute pulse waveforms
                        add     pulse1_val, PULSE_STEP              'increment pulse counter  
                        cmp     pulse1_end, pulse1_val   wc         'if pulse end reached
        if_nc           jmp     #:ldone_pulse2                      '{
                        add     pulse1_init, #2                     '   reduce low time by 1 sample
                        mov     pulse1_val, pulse1_init             '   adopt new pulse timing
                        
                        cmp     pulse1_init, LONG_HALF wc           '   if low time has reached zero
        if_c            jmp     #:ldone_pulse2                      '   {
                        sub     pulse1_end, PULSE_MOD_STEP          '      reduce high time by 1 sample
                        add     pulse1_base_init, PULSE_MOD_STEP
                        mov     pulse1_init, LONG_HALF              '      set low time back to starting width
                        sub     pulse1_init, pulse1_base_init

                        mov     pulse2_init, LONG_HALF
                        sub     pulse2_init, PULSE2_INIT_START
                                         '
                        cmp     pulse1_end, LONG_HALF   wc          '      if high time reached zero
                                                                    '      {
        if_c            andn    sfx_in_progress, sfx_id_mask        '         clear "in progress" mask (i.e. done)
                                                                    '      }      
                                                                    '   }
                                                                    '} 
:ldone_pulse2           add     pulse2_val, PULSE_STEP              'increment pulse counter
                        cmp     pulse1_end, pulse2_val    wc        'if pulse end reached   
        if_c            mov     pulse2_val, pulse2_init             '   reload pulse 2

                        '------------------------------------
                        'Determine output waveform state
                        '------------------------------------
                        
                        'Output high if PULSE_BIT is high in either pulse generator
                        mov     AUDIO_channel_1, #00

                        cmp     pulse2_val, PULSE_BIT wc
              if_nc     mov     AUDIO_channel_1,#$ff
         
                        cmp     pulse1_val, PULSE_BIT wc
              if_nc     mov     AUDIO_channel_1,#$ff


                        'Add edge emphasis if transitioning
                        cmp     previous_pulse_level, #0  wz
                  if_nz jmp     #:ldone_check_falling
                        cmp     AUDIO_channel_1, #$ff  wz
                   if_z or      flags, #FLAG_RISING_EDGE
                        jmp     #:ldone_bypass

:ldone_check_falling    cmp     AUDIO_channel_1, #0   wz
                   if_z or      flags, #FLAG_FALLING_EDGE  

:ldone_bypass


                        '================================================================================
                        'PLAY: Start Game 
                        '================================================================================
                        shl     sfx_id_mask, #1 
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:sgame_bypass

                       
                        '------------------------------------
                        'Do null interval
                        '------------------------------------
                        
                        cmp     SGAME_angle, #$80  wc
                 if_nc  jmp     #:sgame_wave
                        add     SGAME_angle, #1
                        mov     y, #0
                        jmp     #:sgame_done
                        

                        '------------------------------------
                        'Do wave cycle interval
                        '------------------------------------
:sgame_wave
                        'Get sine of current angle
                        add     SGAME_angle, SGAME_ANGLE_STEP
                        mov     sin, SGAME_angle
                        shr     sin, #10
                        call    #getsin
                        add     sin, HALF_MAX_SIN
                        shr     sin, #9

                        'Scale by current scaling factor 
                        mov     x, sin
                        mov     y, SGAME_scale
                        shr     y, #5
                        cmp     y, #5
                   if_z neg     y, y
                   if_z add     y, #9
                        call    #mult

                        'Truncate
                        and     y, #$ff

                        'Update
                        cmp     SGAME_angle, SGAME_ANGLE_END wc
                 if_c   jmp     #:sgame_done
                        mov     SGAME_angle, #0
                        add     SGAME_scale, #1
                        mov     x,SGAME_scale
                        and     x, #$1f
                        cmp     x, #$12   wz
                 if_nz  jmp     #:sgame_done
                        add     SGAME_scale, #$20
                        andn    SGAME_scale, #$1f
                        cmp     SGAME_scale, #$1A0  wz
                 if_z   andn    sfx_in_progress, sfx_id_mask
                        
:sgame_done  
                        '------------------------------------
                        'Set PWM output
                        '------------------------------------
                        shl     y,#2
                        mov     AUDIO_channel_1, y

:sgame_bypass

                        '================================================================================
                        'Audio Mixing
                        '================================================================================  
:do_mixing

                        '------------------------------------
                        'Mix audio channels
                        '------------------------------------
                        mov     r2, AUDIO_channel_0
                        add     r2, AUDIO_channel_1                                                     
                        shl     r2, #18
                        add     r2, LONG_HALF

                        '------------------------------------
                        'Add edge emphasis to increase square wave rise times
                        '------------------------------------
                        
                        'Add edge emphasis if transitioning
                        test    flags, #FLAG_RISING_EDGE   wz
                  if_nz add     r2, EDGE_EMPHASIS_INIT                          'add edge emphasis
                        test    flags, #FLAG_FALLING_EDGE  wz
                  if_nz sub     r2, EDGE_EMPHASIS_INIT                          'add edge emphasis
                        
                        '------------------------------------
                        'Output wave data
                        '------------------------------------
                        mov     frqa, r2

                        

                        '================================================================================
                        'Done main loop
                        '================================================================================
                        
                        jmp     #:sample_sync


'------------------------------------ 
' Get sine/cosine
'
'       quadrant:    1            2            3            4
'          angle:    $0000..$07FF $0800..$0FFF $1000..$17FF $1800..$1FFF
'    table index:    $0000..$07FF $0800..$0001 $0000..$07FF $0800..$0001
'         mirror:    +offset      -offset      +offset      -offset
'           flip:    +sample      +sample      -sample      -sample
'
' on entry: sin[12..0] holds angle (0 to just under 360)
' on exit: sin holds signed value ranging from $0000FFFF ('1') to
' $FFFF0001 ('-1')
'------------------------------------ 
getcos                  add     sin,sin_90      'for cosine, add 90
getsin                  test    sin,sin_90 wc   'get quadrant 2|4 into c
                        test    sin,sin_180 wz  'get quadrant 3|4 into nz
                        negc    sin,sin         'if quadrant 2|4, negate offset
                        or      sin,sin_table   'or in sin table address >> 1
                        shl     sin,#1          'shift left to get final word address
                        rdword  sin,sin         'read word sample from $E000 to $F000
                        negnz   sin,sin         'if quadrant 3|4, negate sample
getsin_ret
getcos_ret              ret                     '39..54 clocks
                                                '(variance due to HUB sync on RDWORD)
sin_90                  long    $0800
sin_180                 long    $1000
sin_table               long    $E000 >> 1      'sine table base shifted right
sin                     long    0
HALF_MAX_SIN            long    $FFFF                               


'------------------------------------
'Multiply                                    
'------------------------------------
' Multiply x[15..0] by y[15..0] (y[31..16] must be 0)
' on exit, product in y[31..0]
'------------------------------------
mult                    shl x,#16               'get multiplicand into x[31..16]
                        mov t,#16               'ready for 16 multiplier bits
                        shr y,#1 wc             'get initial multiplier bit into c
:loop
                        if_c add y,x wc         'conditionally add multiplicand into product
                        rcr y,#1 wc             'get next multiplier bit into c.
                                                ' while shift product
                        djnz t,#:loop           'loop until done
mult_ret                ret                     'return with product in y[31..0]

'------------------------------------
'Initialized Data                                      
'------------------------------------
_PIN__AUDIO             long    PIN__AUDIO

SAMPLE_INTERVAL_HYDRA   long   (CLOCK_RATE_HYDRA / AUDIO_SAMPLE_RATE)
SAMPLE_INTERVAL_HYBRID  long   (CLOCK_RATE_HYBRID / AUDIO_SAMPLE_RATE)    

CTRAVAL                 long   %00110 << 26 + 7  'DUTY single-ended, APIN=7        '
LONG_HALF               long   $7FFF_FFFF
PULSE_BIT               long   $8000_0000
PULSE1_INIT_START       long   44 * 8
PULSE1_END_START        long   14 * 8
PULSE2_INIT_START       long   162 * 8
PULSE2_END_START        long   12 * 8
PULSE_MOD_STEP          long   10
PULSE_STEP              long   8

SOUND__LEVEL_DONE_ID    long   SOUND__LEVEL_DONE

LFSR_32_TAPS            long   $800000c2

EDGE_EMPHASIS_INIT      long   $0fffffff

SGAME_ANGLE_STEP        long   $0000f587
SGAME_ANGLE_END         long   $00800000

'------------------------------------
'Uninitialized Data
'------------------------------------
                          
r1                        res     1
r2                        res     1
sfx_id_mask               res     1
flags                     res     1
sample_interval_ticks     res     1

x                         res     1    ' for multiply
y                         res     1    ' for multiply
t                         res     1    ' for multiply

previous_pulse_level      res     1

start_request_previous    res     1
start_request_current     res     1
start_request_actions     res     1
sfx_in_progress           res     1
stop_request_current      res     1

p_sfx_start_flags         res     1    
p_sfx_stop_flags          res     1
sfx_playing_mask          res     1

next_sample_start_time    res     1

random_number             res     1

pulse1_base_init          res     1
pulse1_init               res     1
pulse2_init               res     1
pulse1_end                res     1
pulse2_end                res     1
pulse1_val                res     1
pulse2_val                res     1

'Audio channels
AUDIO_channel_0           res     1  
AUDIO_channel_1           res     1    

'START parameters
SGAME_angle               res     1
SGAME_scale               res     1

                          fit                 